package opensteer;

import javax.vecmath.Vector3f;

// Converted from the original C++ by JIST
// Use this as a basis for transcribing into Java... (good luck!)




// ----------------------------------------------------------------------------
//
//
// OpenSteer -- Steering Behaviors for Autonomous Characters
//
// Copyright (c) 2002-2003, Sony Computer Entertainment America
// Original author: Craig Reynolds <craig_reynolds@playstation.sony.com>
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//
//
// ----------------------------------------------------------------------------
//
//
// LocalSpace: a local coordinate system for 3d space
//
// Provide functionality such as transforming from local space to global
// space and vice versa.  Also regenerates a valid space from a perturbed
// "forward vector" which is the basis of abnstract vehicle turning.
//
// These are comparable to a 4x4 homogeneous transformation matrix where the
// 3x3 (R) portion is constrained to be a pure rotation (no shear or scale).
// The rows of the 3x3 R matrix are the basis vectors of the space.  They are
// all constrained to be mutually perpendicular and of unit length.  The top
// ("x") row is called "side", the middle ("y") row is called "up" and the
// bottom ("z") row is called forward.  The translation vector is called
// "position".  Finally the "homogeneous column" is always [0 0 0 1].
//
//     [ R R R  0 ]      [ Sx Sy Sz  0 ]
//     [ R R R  0 ]      [ Ux Uy Uz  0 ]
//     [ R R R  0 ]  ->  [ Fx Fy Fz  0 ]
//     [          ]      [             ]
//     [ T T T  1 ]      [ Tx Ty Tz  1 ]
//
// This file defines three classes:
//   AbstractLocalSpace:  pure virtual interface
//   LocalSpaceMixin:     mixin to layer LocalSpace functionality on any base
//   LocalSpace:          a concrete object (can be instantiated)
//
// 06-05-02 cwr: created 
//
//
// ----------------------------------------------------------------------------
public interface AbstractLocalSpace
{
    // accessors (get and set) for side, up, forward and position
    public Vector3f side ();

    public void setSide (Vector3f s);

    public Vector3f up ();

    public void setUp (Vector3f u);

    public Vector3f forward ();

    public void setForward (Vector3f f);

    public Vector3f position ();

    public void setPosition (Vector3f p);


    // use right-(or left-)handed coordinate space
    public boolean rightHanded ();


    // reset transform to identity
    public void resetLocalSpace ();


    // transform a direction in global space to its equivalent in local space
    public Vector3f localizeDirection (Vector3f globalDirection);


    // transform a point in global space to its equivalent in local space
    public Vector3f localizePosition (Vector3f globalPosition);


    // transform a point in local space to its equivalent in global space
    public Vector3f globalizePosition (Vector3f localPosition);


    // transform a direction in local space to its equivalent in global space
    public Vector3f globalizeDirection (Vector3f localDirection);


    // set "side" basis vector to normalized cross product of forward and up
    public void setUnitSideFromForwardAndUp ();


    // regenerate the orthonormal basis vectors given a new forward
    // (which is expected to have unit length)
    public void regenerateOrthonormalBasisUF (Vector3f newUnitForward);


    // for when the new forward is NOT of unit length
    public void regenerateOrthonormalBasis (Vector3f newForward);


    // for supplying both a new forward and and new up
    public void regenerateOrthonormalBasis (Vector3f newForward, Vector3f newUp);


    // rotate 90 degrees in the direction implied by rightHanded()
    public Vector3f localRotateForwardToSide (Vector3f v);

    public Vector3f globalRotateForwardToSide (Vector3f globalForward);
}

