package opensteer;

import javax.vecmath.Color3f;
import javax.vecmath.Vector3f;

// Converted from the original C++ by JIST
// Use this as a basis for transcribing into Java... (good luck!)

public abstract class Annotation
{
    public static final Color3f gBlack   = new Color3f (0, 0, 0);
    public static final Color3f gWhite   = new Color3f (1, 1, 1);

    public static final Color3f gRed     = new Color3f (1, 0, 0);
    public static final Color3f gYellow  = new Color3f (1, 1, 0);
    public static final Color3f gGreen   = new Color3f (0, 1, 0);
    public static final Color3f gCyan    = new Color3f (0, 1, 1);
    public static final Color3f gBlue    = new Color3f (0, 0, 1);
    public static final Color3f gMagenta = new Color3f (1, 0, 1);

    public static final Color3f gOrange = new Color3f (1, 0.5f, 0);

    static Color3f grayColor (float g) {return new Color3f (g, g, g);}

    public static final Color3f gGray10 = grayColor (0.1f);
    public static final Color3f gGray20 = grayColor (0.2f);
    public static final Color3f gGray30 = grayColor (0.3f);
    public static final Color3f gGray40 = grayColor (0.4f);
    public static final Color3f gGray50 = grayColor (0.5f);
    public static final Color3f gGray60 = grayColor (0.6f);
    public static final Color3f gGray70 = grayColor (0.7f);
    public static final Color3f gGray80 = grayColor (0.8f);
    public static final Color3f gGray90 = grayColor (0.9f);


    private int trailVertexCount;
    private int trailIndex;
    private float trailDuration;
    private float trailSampleInterval;
    private float trailLastSampleTime;
    private int trailDottedPhase;
    private Vector3f curPosition;
    private Vector3f trailVertices;
    private String trailFlags;

    private Annotation()
    {
        trailVertices = null;
        trailFlags = null;

        // xxx I wonder if it makes more sense to NOT do this here, see if the
        // xxx vehicle class calls it to set custom parameters, and if not, set
        // xxx these default parameters on first call to a "trail" function.  The
        // xxx issue is whether it is a problem to allocate default-sized arrays
        // xxx then to free them and allocate new ones
        setTrailParameters(5, 100); // 5 seconds with 100 points along the trail
    }

    // ------------------------------------------------------------------------
    // trails / streamers
    //
    // these routines support visualization of a vehicle's recent path
    //
    // XXX conceivable trail/streamer should be a separate class,
    // XXX Annotation would "has-a" one (or more))

    // record a position for the current time, called once per update
    public abstract void recordTrailVertex(float currentTime, Vector3f position);

    // draw the trail as a dotted line, fading away with age
    public abstract void drawTrail();

    // set trail parameters: the amount of time it represents and the
    // number of samples along its length.  re-allocates internal buffers.
    public abstract void setTrailParameters(float duration, int vertexCount);

    // forget trail history: used to prevent long streaks due to teleportation
    public abstract void clearTrailHistory();

    // ------------------------------------------------------------------------
    // drawing of lines, circles and (filled) disks to annotate steering
    // behaviors.  When called during SteerTest's simulation update phase,
    // these functions call a "deferred draw" routine which buffer the
    // arguments for use during the redraw phase.
    //
    // note: "circle" means unfilled
    //       "disk" means filled
    //       "XZ" means on a plane parallel to the X and Z axes (perp to Y)
    //       "3d" means the circle is perpendicular to the given "axis"
    //       "segments" is the number of line segments used to draw the circle

    // draw an opaque colored line segment between two locations in space
    public abstract void annotationLine(Vector3f startPoint, Vector3f endPoint, Vector3f color);

    // draw a circle on the XZ plane
    public void annotationXZCircle(float radius, Vector3f center, Vector3f color, int segments)
    {
        annotationXZCircleOrDisk(radius, center, color, segments, false);
    }

    // draw a disk on the XZ plane
    public void annotationXZDisk(float radius, Vector3f center, Vector3f color, int segments)
    {
        annotationXZCircleOrDisk(radius, center, color, segments, true);
    }

    // draw a circle perpendicular to the given axis
    public void annotation3dCircle(float radius, Vector3f center, Vector3f axis, Vector3f color, int segments)
    {
        annotation3dCircleOrDisk(radius, center, axis, color, segments, false);
    }

    // draw a disk perpendicular to the given axis
    public void annotation3dDisk(float radius, Vector3f center, Vector3f axis, Vector3f color, int segments)
    {
        annotation3dCircleOrDisk(radius, center, axis, color, segments, true);
    }

    //

    // ------------------------------------------------------------------------
    // support for annotation circles

    public void annotationXZCircleOrDisk(float radius, Vector3f center, Vector3f color, int segments, boolean filled)
    {
        annotationCircleOrDisk(radius, new Vector3f (0,0,0), center, color, segments, filled, false); // "not in3d" -> on XZ plane
    }

    public void annotation3dCircleOrDisk(float radius, Vector3f center, Vector3f axis, Vector3f color, int segments, boolean filled)
    {
        annotationCircleOrDisk(radius, axis, center, color, segments, filled, true); // "in3d"
    }

    public abstract void annotationCircleOrDisk(float radius, Vector3f axis, Vector3f center, Vector3f color, int segments, boolean filled, boolean in3d);
}
