//Converted from the original C++ by JIST
//Use this as a basis for transcribing into Java... (good luck!)

//----------------------------------------------------------------------------
//
//
//OpenSteer -- Steering Behaviors for Autonomous Characters
//
//Copyright (c) 2002-2003, Sony Computer Entertainment America
//Original author: Craig Reynolds <craig_reynolds@playstation.sony.com>
//
//Permission is hereby granted, free of charge, to any person obtaining a
//copy of this software and associated documentation files (the "Software"),
//to deal in the Software without restriction, including without limitation
//the rights to use, copy, modify, merge, publish, distribute, sublicense,
//and/or sell copies of the Software, and to permit persons to whom the
//Software is furnished to do so, subject to the following conditions:
//
//The above copyright notice and this permission notice shall be included in
//all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
//THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
//FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
//DEALINGS IN THE SOFTWARE.
//
//
//----------------------------------------------------------------------------
//
//
//LocalSpace: a local coordinate system for 3d space
//
//Provide functionality such as transforming from local space to global
//space and vice versa.  Also regenerates a valid space from a perturbed
//"forward vector" which is the basis of abnstract vehicle turning.
//
//These are comparable to a 4x4 homogeneous transformation matrix where the
//3x3 (R) portion is constrained to be a pure rotation (no shear or scale).
//The rows of the 3x3 R matrix are the basis vectors of the space.  They are
//all constrained to be mutually perpendicular and of unit length.  The top
//("x") row is called "side", the middle ("y") row is called "up" and the
//bottom ("z") row is called forward.  The translation vector is called
//"position".  Finally the "homogeneous column" is always [0 0 0 1].
//
//  [ R R R  0 ]      [ Sx Sy Sz  0 ]
//  [ R R R  0 ]      [ Ux Uy Uz  0 ]
//  [ R R R  0 ]  ->  [ Fx Fy Fz  0 ]
//  [          ]      [             ]
//  [ T T T  1 ]      [ Tx Ty Tz  1 ]
//
//This file defines three classes:
//AbstractLocalSpace:  pure virtual interface
//LocalSpaceMixin:     mixin to layer LocalSpace functionality on any base
//LocalSpace:          a concrete object (can be instantiated)
//
//06-05-02 cwr: created 
//
//
//----------------------------------------------------------------------------

package opensteer;

import javax.vecmath.Matrix4f;
import javax.vecmath.Vector3f;

import common.SimulatedObject;

public class LocalSpace
{
    protected Vector3f side = new Vector3f ();
    protected Vector3f up = new Vector3f ();
    protected Vector3f forward = new Vector3f ();
    protected Vector3f position = new Vector3f ();

    // These were originally in Vec3, but since I've replaced that by Vector3f, here seemed a good alternative home.
    
    public static final Vector3f ZERO    = new Vector3f (0, 0, 0);
    public static final Vector3f UP      = new Vector3f (0, 0, 1);
    public static final Vector3f FORWARD = new Vector3f (1, 0, 0);
    public static final Vector3f SIDE    = new Vector3f (0, -1, 0);
    
    // accessors (get and set) for side, up, forward and position
    public Vector3f side()
    {
        return side;
    }

    public Vector3f up()
    {
        return up;
    }

    public Vector3f forward()
    {
        return forward;
    }

    public Vector3f position()
    {
        return position;
    }

    public void setSide(Vector3f s)
    {
        side.set (s);
    }

    public void setUp(Vector3f u)
    {
        up.set (u);
    }

    public void setForward(Vector3f f)
    {
        forward.set (f);
    }

    public void setPosition(Vector3f p)
    {
        position.set (p);
    }

    public void setSide(float x, float y, float z)
    {
        side.set(x, y, z);
    }

    public void setUp(float x, float y, float z)
    {
        up.set(x, y, z);
    }

    public void setForward(float x, float y, float z)
    {
        forward.set(x, y, z);
    }

    public void setPosition(float x, float y, float z)
    {
        position.set(x, y, z);
    }
    
    public void getTransform (Matrix4f m)
    {
        m.setColumn (1, side.x, side.y, side.z, 0);
        m.setColumn (2, up.x, up.y, up.z, 0);
        m.setColumn (0, forward.x, forward.y, forward.z, 0);
        m.setColumn (3, position.x, position.y, position.z, 1);
    }

    // ------------------------------------------------------------------------
    // Global compile-time switch to control handedness/chirality: should
    // LocalSpace use a left- or right-handed coordinate system?  This can be
    // overloaded in derived types (e.g. vehicles) to change handedness.

    public boolean rightHanded()
    {
        return true;
    }

    // ------------------------------------------------------------------------
    // constructors

    public LocalSpace()
    {
        resetLocalSpace();
    }

    public LocalSpace (Vector3f Side, Vector3f Up, Vector3f Forward, Vector3f Position)
    {
        side.set (Side);
        up.set (Up);
        forward.set (Forward);
        position.set (Position);
    }

    public LocalSpace(Vector3f Up, Vector3f Forward, Vector3f Position)
    {
        up.set (Up);
        forward.set (Forward);
        position.set (Position);
        setUnitSideFromForwardAndUp();
    }

    // ------------------------------------------------------------------------
    // reset transform: set local space to its identity state, equivalent to a
    // 4x4 homogeneous transform like this:
    //
    //     [ X 0 0 0 ]
    //     [ 0 1 0 0 ]
    //     [ 0 0 1 0 ]
    //     [ 0 0 0 1 ]
    //
    // where X is 1 for a left-handed system and -1 for a right-handed system.

    public void resetLocalSpace()
    {
        forward.set(FORWARD);
        side.set (SIDE); // side = localRotateForwardToSide(forward);
        up.set(UP);
        position.set(0, 0, 0);
    }

    // ------------------------------------------------------------------------
    // transform a direction in global space to its equivalent in local space

    public Vector3f localizeDirection(Vector3f globalDirection)
    {
        // dot offset with local basis vectors to obtain local coordiantes
        return new Vector3f (globalDirection.dot(side), globalDirection.dot(up), globalDirection.dot(forward));
    }

    // ------------------------------------------------------------------------
    // transform a point in global space to its equivalent in local space

    public Vector3f localizePosition(Vector3f globalPosition)
    {
        // global offset from local origin
        Vector3f globalOffset = new Vector3f (globalPosition); globalOffset.sub (position);

        // dot offset with local basis vectors to obtain local coordiantes
        return localizeDirection(globalOffset);
    }

    // ------------------------------------------------------------------------
    // transform a point in local space to its equivalent in global space

    public Vector3f globalizePosition(Vector3f localPosition)
    {
        Vector3f v = new Vector3f (position); v.add (globalizeDirection(localPosition));
        return v;
    }

    // ------------------------------------------------------------------------
    // transform a direction in local space to its equivalent in global space

    public Vector3f globalizeDirection(Vector3f localDirection)
    {
        Vector3f v1 = new Vector3f (side);    v1.scale (localDirection.x);
        Vector3f v2 = new Vector3f (up);      v2.scale (localDirection.y);
        Vector3f v3 = new Vector3f (forward); v3.scale (localDirection.z);
        
        v1.add (v2); v1.add (v3);
        
        return v1;
    }

    // ------------------------------------------------------------------------
    // set "side" basis vector to normalized cross product of forward and up

    public void setUnitSideFromForwardAndUp()
    {
        // derive new unit side basis vector from forward and up
        if (rightHanded())
            side.cross(forward, up);
        else
            side.cross(up, forward);
            
        side.normalize();
    }

    // ------------------------------------------------------------------------
    // regenerate the orthonormal basis vectors given a new forward
    // (which is expected to have unit length)

    public void regenerateOrthonormalBasisUF(Vector3f newUnitForward)
    {
        forward.set (newUnitForward);

        // derive new side basis vector from NEW forward and OLD up
        setUnitSideFromForwardAndUp();

        // derive new Up basis vector from new Side and new Forward
        // (should have unit length since Side and Forward are
        // perpendicular and unit length)
        if (rightHanded())
            up.cross(side, forward);
        else
            up.cross(forward, side);
    }

    // for when the new forward is NOT know to have unit length

    public void regenerateOrthonormalBasis(Vector3f newForward)
    {
        Vector3f v = new Vector3f (newForward); v.normalize();
        regenerateOrthonormalBasisUF (v);
    }

    // for supplying both a new forward and and new up

    public void regenerateOrthonormalBasis(Vector3f newForward, Vector3f newUp)
    {
        up.set (newUp);
        Vector3f v = new Vector3f (newForward); v.normalize();
        regenerateOrthonormalBasis (v);
    }

    // ------------------------------------------------------------------------
    // rotate, in the canonical direction, a vector pointing in the
    // "forward" (+Z) direction to the "side" (+/-X) direction

    public Vector3f localRotateForwardToSide(Vector3f v)
    {
        return new Vector3f (rightHanded() ? -v.y : +v.y, v.x, v.z);
    }

    // not currently used, just added for completeness

    public Vector3f globalRotateForwardToSide(Vector3f globalForward)
    {
        final Vector3f localForward = localizeDirection(globalForward);
        final Vector3f localSide = localRotateForwardToSide(localForward);
        return globalizeDirection(localSide);
    }
}
