package opensteer;

import javax.vecmath.Vector3f;

// Converted from the original C++ by JIST
// Use this as a basis for transcribing into Java... (good luck!)

// ----------------------------------------------------------------------------
// PolylinePathway: a simple implementation of the Pathway protocol.  The path
// is a "polyline" a series of line segments between specified points.  A
// radius defines a volume for the path which is the union of a sphere at each
// point and a cylinder along each segment.

public class PolylinePathway implements Pathway
{
    public int pointCount;
    public Vector3f[] points;
    public float radius;
    public boolean cyclic;

    private float segmentLength;
    private float segmentProjection;
    private Vector3f segmentNormal;
    private Vector3f chosen;
    private float[] lengths;
    private Vector3f[] normals;
    private float totalPathLength;

    // assessor for total path length;
    public float getTotalPathLength()
    {
        return totalPathLength;
    }

    // ----------------------------------------------------------------------------
    // construct a PolylinePathway given the number of points (vertices),
    // an array of points, and a path radius.

    public PolylinePathway(int pointCount, Vector3f[] points, float radius, boolean cyclic)
    {
        // set data members, allocate arrays
        this.radius = radius;
        this.cyclic = cyclic;
        this.pointCount = pointCount;
        this.totalPathLength = 0;
        
        if (cyclic)
            pointCount++;
            
        lengths = new float[pointCount];
        this.points = new Vector3f[pointCount];
        normals = new Vector3f[pointCount];

        // loop over all points
        for (int i = 0; i < pointCount; i++)
        {
            // copy in point locations, closing cycle when appropriate
            final boolean closeCycle = cyclic && (i == pointCount - 1);
            final int j = closeCycle ? 0 : i;
            this.points[i] = points[j];

            // for the end of each segment
            if (i > 0)
            {
                // compute the segment length
                // normals[i] = points[i] - points[i - 1];
                normals[i] = new Vector3f(points[i]); normals[i].sub (points [i-1]);
                lengths[i] = normals[i].length();

                // find the normalized vector parallel to the segment
                normals[i].scale (1 / lengths[i]);

                // keep running total of segment lengths
                totalPathLength += lengths[i];
            }
        }
    }

    /** Given an arbitrary point ("A"), finds the nearest point ("P") on this
     * path.  Also sets the path tangent at P and returns a flag indicating 
     *  whether or not A was already inside the Pathway "tube".
     */
    public float mapPointToPath(Vector3f point, Vector3f onPath, Vector3f tangent)
    {
        float d;
        float minDistance = Float.MAX_VALUE;

        // loop over all segments, find the one nearest to the given point
        for (int i = 1; i < pointCount; i++)
        {
            segmentLength = lengths[i];
            segmentNormal = normals[i];
            d = pointToSegmentDistance(point, points[i - 1], points[i]);

            if (d < minDistance)
            {
                minDistance = d;
                onPath.set (chosen);
                tangent.set (segmentNormal);
            }
        }
        // was original point inside Pathway "tube"?
        return Vector3fUtil.distance(onPath, point);
    }

    // ----------------------------------------------------------------------------
    // given an arbitrary point, convert it to a distance along the path

    public float mapPointToPathDistance(Vector3f point)
    {
        float d;
        float minDistance = 3.40282347e+38F;
        float segmentLengthTotal = 0;
        float pathDistance = 0;

        for (int i = 1; i < pointCount; i++)
        {
            segmentLength = lengths[i];
            segmentNormal = normals[i];
            d = pointToSegmentDistance(point, points[i - 1], points[i]);
            if (d < minDistance)
            {
                minDistance = d;
                pathDistance = segmentLengthTotal + segmentProjection;
            }
            segmentLengthTotal += segmentLength;
        }

        // return distance along path of onPath point
        return pathDistance;
    }

    // ----------------------------------------------------------------------------
    // given a distance along the path, convert it to a point on the path

    public Vector3f mapPathDistanceToPoint(float pathDistance)
    {
        // clip or wrap given path distance according to cyclic flag
        float remaining = pathDistance;
        if (cyclic)
        {
            remaining = pathDistance % totalPathLength;
        }
        else
        {
            if (pathDistance < 0)
                remaining = 0;
            if (pathDistance > totalPathLength)
                remaining = totalPathLength;
        }

        // step through segments, subtracting off segment lengths until
        // locating the segment that contains the original pathDistance.
        // Interpolate along that segment to find 3d point value to return.
        Vector3f result = new Vector3f();
        for (int i = 1; i < pointCount; i++)
        {
            segmentLength = lengths[i];
            if (segmentLength < remaining)
            {
                remaining -= segmentLength;
            }
            else
            {
                float ratio = remaining / segmentLength;
                result.interpolate (points[i - 1], points[i], ratio);
                break;
            }
        }
        return result;
    }

    // ----------------------------------------------------------------------------
    // computes distance from a point to a line segment 
    //
    // (I considered moving this to the vector library, but its too
    // tangled up with the internal state of the PolylinePathway instance)

    private float pointToSegmentDistance(Vector3f point, Vector3f ep0, Vector3f ep1)
    {
        // convert the test point to be "local" to ep0
        Vector3f local = new Vector3f (point); local.sub (ep0);

        // find the projection of "local" onto "segmentNormal"
        segmentProjection = segmentNormal.dot(local);

        // handle boundary cases: when projection is not on segment, the
        // nearest point is one of the endpoints of the segment
        if (segmentProjection < 0)
        {
            chosen = ep0;
            segmentProjection = 0;
            return Vector3fUtil.distance(point, ep0);
        }
        if (segmentProjection > segmentLength)
        {
            chosen = ep1;
            segmentProjection = segmentLength;
            return Vector3fUtil.distance(point, ep1);
        }

        // otherwise nearest point is projection point on segment
        chosen = new Vector3f (segmentNormal); chosen.scale (segmentProjection);
        chosen.add (ep0);
        return Vector3fUtil.distance(point, chosen);
    }

	public Vector3f[] getPoints() {
    	return points;
    }

	public float getRadius() {
    	return radius;
    }
}
