package opensteer;

import javax.vecmath.Color3f;
import javax.vecmath.Vector3f;

import common.SimulatedObject;

// Converted from the original C++ by JIST
// Use this as a basis for transcribing into Java... (good luck!)

//----------------------------------------------------------------------------
//
//
//OpenSteer -- Steering Behaviors for Autonomous Characters
//
//Copyright (c) 2002-2003, Sony Computer Entertainment America
//Original author: Craig Reynolds <craig_reynolds@playstation.sony.com>
//
//Permission is hereby granted, free of charge, to any person obtaining a
//copy of this software and associated documentation files (the "Software"),
//to deal in the Software without restriction, including without limitation
//the rights to use, copy, modify, merge, publish, distribute, sublicense,
//and/or sell copies of the Software, and to permit persons to whom the
//Software is furnished to do so, subject to the following conditions:
//
//The above copyright notice and this permission notice shall be included in
//all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
//THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
//FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
//DEALINGS IN THE SOFTWARE.
//
//
//----------------------------------------------------------------------------
//
//
//SimpleVehicle
//
//A steerable point mass with a velocity-aligned local coordinate system.
//SimpleVehicle is useful for developing prototype vehicles in SteerTest, it
//is the base class for vehicles in the PlugIns supplied with OpenSteer.
//Note that SimpleVehicle is provided only as sample code.  Your application
//can use the OpenSteer library without using SimpleVehicle, as long as you
//implement the AbstractVehicle protocol.
//
//SimpleVehicle makes use of the "mixin" concept from OOP.  To quote
//"Mixin-Based Programming in C++" a clear and helpful paper by Yannis
//Smaragdakis and Don Batory (http://citeseer.nj.nec.com/503808.html):
//
//  ...The idea is simple: we would like to specify an extension without
//  predetermining what exactly it can extend. This is equivalent to
//  specifying a subclass while leaving its superclass as a parameter to be
//  determined later. The benefit is that a single class can be used to
//  express an incremental extension, valid for a variety of classes...
// 
//In OpenSteer, vehicles are defined by an interface: an abstract base class
//called AbstractVehicle.  Implementations of that interface, and related
//functionality (like steering behaviors and vehicle physics) are provided as
//template-based mixin classes.  The intent of this design is to allow you to
//reuse OpenSteer code with your application's own base class.
//
//01-29-03 cwr: created
//
//
//----------------------------------------------------------------------------

public class SimpleVehicle extends LocalSpace implements AbstractVehicle, SimulatedObject
{
    public int serialNumber;
    public int serialNumberCounter;

    private float mass;
    private float radius;
    private float speed;
    private float maxForce;
    private float maxSpeed;
    protected Vector3f smoothedAcceleration = new Vector3f ();

    // ----------------------------------------------------------------------------
    // constructor

    public SimpleVehicle()
    {
        // set inital state
        reset();

        // maintain unique serial numbers
        serialNumber = serialNumberCounter++;
    }

    // reset vehicle state
    public void reset()
    {
        // reset LocalSpace state
        resetLocalSpace();

        // reset SteerLibraryMixin state
        // (XXX this seems really fragile, needs to be redesigned XXX)
        // SimpleVehicle_3.reset();

        setMass(1); // mass (defaults to 1 so acceleration=force)
        setSpeed(0); // speed along Forward direction.

        setRadius(0.5f); // size of bounding sphere

        setMaxForce(0.1f); // steering force is clipped to this magnitude
        setMaxSpeed(1.0f); // velocity is clipped to this magnitude

        smoothedAcceleration.set (ZERO);
    }

    // get/set mass
    public float mass()
    {
        return mass;
    }

    public float setMass(float m)
    {
        return mass = m;
    }

    // get velocity of vehicle
    public Vector3f velocity()
    {
        // return localSpace.forward() * speed;
        Vector3f v = new Vector3f (forward()); v.scale (speed); return v;
    }

    // get/set speed of vehicle  (may be faster than taking mag of velocity)
    public float speed()
    {
        return speed;
    }

    public float setSpeed(float s)
    {
        return speed = s;
    }

    // size of bounding sphere, for obstacle avoidance, etc.
    public float radius()
    {
        return radius;
    }

    public float setRadius(float m)
    {
        return radius = m;
    }

    // get/set maxForce
    public float maxForce()
    {
        return maxForce;
    }

    public float setMaxForce(float mf)
    {
        return maxForce = mf;
    }

    // get/set maxSpeed
    public float maxSpeed()
    {
        return maxSpeed;
    }

    public float setMaxSpeed(float ms)
    {
        return maxSpeed = ms;
    }

    // set a random "2D" heading: set local Up to global Y, then effectively
    // rotate about it by a random angle (pick random forward, derive side).
    public void randomizeHeadingOnXZPlane()
    {
        setUp (LocalSpace.UP);
        setForward(Vector3fUtil.randomUnitVectorOnXZPlane());
        setSide(localRotateForwardToSide(forward()));
    }

    // DL: a more proper way (I believe) to set a random "2D" heading: 
    // leave local up, set local forward to an admixture of previous forward and side.
    public void randomizeHeadingOn2DPlane()
    {
        Vector3f newForwardXZ = Vector3fUtil.randomUnitVectorOnXZPlane();
        Vector3f newForwardFwd = new Vector3f (forward()); newForwardFwd.scale (newForwardXZ.z);
        Vector3f newForwardSide = new Vector3f (side()); newForwardSide.scale (newForwardXZ.x);
        newForwardFwd.add (newForwardSide);
        setForward(newForwardFwd);
        setSide(localRotateForwardToSide(forward()));
    }
//    // set a random "2D" heading: set local Up to global Z, then effectively
//    // rotate about it by a random angle (pick random forward, derive side).
//    public void randomizeHeadingOnXYPlane()
//    {
//        setUp (LocalSpace.UP);
//        setForward(Vector3fUtil.randomUnitVectorOnXYPlane());
//        setSide(localRotateForwardToSide(forward()));
//    }

    // TODO: Need a proper home for these ...
    public static float interpolate (float alpha, float x0, float x1)
    {
        return x0 + ((x1 - x0) * alpha);
    }
    
    public static float clip (float x, float min, float max)
    {
        if (x < min) return min;
        if (x > max) return max;
        return x;
    }

    public static void blendIntoAccumulator (float smoothRate, Vector3f newValue, Vector3f smoothedAccumulator)
    {
        smoothedAccumulator.interpolate (newValue, clip (smoothRate, 0, 1));
    }

    public static int intervalComparison (float x, float lowerBound, float upperBound)
    {
        if (x < lowerBound) return -1;
        if (x > upperBound) return +1;
        return 0;
    }

    // ----------------------------------------------------------------------------
    // adjust the steering force passed to applySteeringForce.
    //
    // allows a specific vehicle class to redefine this adjustment.
    // default is to disallow backward-facing steering at low speed.
    //
    // xxx should the default be this ad-hocery, or no adjustment?
    // xxx experimental 8-20-02

    protected Vector3f adjustRawSteeringForce(Vector3f force, float deltaTime)
    {
        final float maxAdjustedSpeed = 0.1f * maxSpeed();

        if ((speed() > maxAdjustedSpeed) || (force == LocalSpace.ZERO ))
        {
            return force;
        }
        else
        {
            final float range = speed() / maxAdjustedSpeed;
            // const float cosine = interpolate (pow (range, 6), 1.0, -1.0);
            // const float cosine = interpolate (pow (range, 10), 1.0, -1.0);
            // const float cosine = interpolate (pow (range, 20), 1.0, -1.0);
            // const float cosine = interpolate (pow (range, 100), 1.0, -1.0);
            // const float cosine = interpolate (pow (range, 50), 1.0, -1.0);
            final float cosine = interpolate((float) Math.pow(range, 20), 1.0f, -1.0f);
            return Vector3fUtil.limitMaxDeviationAngle(force, cosine, forward());
        }
    }

    // ----------------------------------------------------------------------------
    // xxx experimental 9-6-02
    //
    // apply a given braking force (for a given dt) to our momentum.
    //
    // (this is intended as a companion to applySteeringForce, but I'm not sure how
    // well integrated it is.  It was motivated by the fact that "braking" (as in
    // "capture the flag" endgame) by using "forward * speed * -rate" as a steering
    // force was causing problems in adjustRawSteeringForce.  In fact it made it
    // get NAN, but even if it had worked it would have defeated the braking.
    //
    // maybe the guts of applySteeringForce should be split off into a subroutine
    // used by both applySteeringForce and applyBrakingForce?

    protected void applyBrakingForce(float rate, float deltaTime)
    {
        final float rawBraking = speed() * rate;
        final float clipBraking = ((rawBraking < maxForce()) ? rawBraking : maxForce());

        setSpeed(speed() - (clipBraking * deltaTime));
    }

    // ----------------------------------------------------------------------------
    // apply a given steering force to our momentum,
    // adjusting our orientation to maintain velocity-alignment.

    protected void applySteeringForce(Vector3f force, float elapsedTime)
    {

        final Vector3f adjustedForce = force; adjustRawSteeringForce(force, elapsedTime);

        // enforce limit on magnitude of steering force
        final Vector3f clippedForce = Vector3fUtil.truncateLength (adjustedForce, maxForce ());

        // compute acceleration and velocity
        // Vector3f newAcceleration = (clippedForce / mass());
        Vector3f newAcceleration = new Vector3f (clippedForce); newAcceleration.scale (1 / mass());
        Vector3f newVelocity = new Vector3f (velocity());

        // damp out abrupt changes and oscillations in steering acceleration
        // (rate is proportional to time step, then clipped into useful range)
        final float smoothRate = clip(9 * elapsedTime, 0.15f, 0.4f);
        blendIntoAccumulator(smoothRate, newAcceleration, smoothedAcceleration);

        // Euler integrate (per frame) acceleration into velocity
        // newVelocity += smoothedAcceleration * elapsedTime;
        Vector3f v = new Vector3f (smoothedAcceleration); v.scale (elapsedTime);
        newVelocity.add (v);

        // enforce speed limit
        newVelocity = Vector3fUtil.truncateLength (newVelocity, maxSpeed());

        // update Speed
        setSpeed(newVelocity.length());

        // Euler integrate (per frame) velocity into position
        // setPosition(position() + (newVelocity * elapsedTime));
        Vector3f w = new Vector3f (newVelocity); w.scale (elapsedTime); w.add (position ());
        setPosition (w);

        // adjust orthonormal basis vectors to be aligned with new velocity
        if (speed() > 0)
        {
            newVelocity.scale (1 / speed());
            regenerateOrthonormalBasisUF (newVelocity);
        }
    }

    // ----------------------------------------------------------------------------
    // draw lines from vehicle's position showing its velocity and acceleration

//    private void annotationVelocityAcceleration(float maxLength)
//    {
//        final float desat = 0.4f;
//        final float aScale = maxLength / maxForce();
//        final float vScale = maxLength / maxSpeed();
//        final Vector3f p = position();
//        final Color3f aColor = new Color3f (desat, desat, 1); // bluish
//        final Color3f vColor = new Color3f (1, desat, 1); // pinkish
//
//        annotationLine(p, p + (velocity() * vScale), vColor);
//        annotationLine(p, p + (smoothedAcceleration * aScale), aColor);
//    }

    // draw lines from vehicle's position showing its velocity and acceleration
//    public void annotationVelocityAcceleration()
//    {
//        annotationVelocityAcceleration(3);
//    }

    // ----------------------------------------------------------------------------
    // predict position of this vehicle at some time in the future
    // (assumes velocity remains constant, hence path is a straight line)
    //
    // XXX Want to encapsulate this since eventually I want to investigate
    // XXX non-linear predictors.  Maybe predictFutureLocalSpace ?
    //
    // XXX move to a vehicle utility mixin?

    public Vector3f predictFuturePosition(float predictionTime)
    {
        // return position() + (velocity() * predictionTime);
        Vector3f v = new Vector3f (velocity()); v.scale (predictionTime); v.add (position()); return v;
    }

	public Vector3f getSmoothedAcceleration() {
		return smoothedAcceleration;
	}
}
