package opensteer;

import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import javax.vecmath.Color3f;
import javax.vecmath.Vector3f;

// Converted from the original C++ by JIST
// Use this as a basis for transcribing into Java... (good luck!)

//----------------------------------------------------------------------------
//
//
//OpenSteer -- Steering Behaviors for Autonomous Characters
//
//Copyright (c) 2002-2003, Sony Computer Entertainment America
//Original author: Craig Reynolds <craig_reynolds@playstation.sony.com>
//
//Permission is hereby granted, free of charge, to any person obtaining a
//copy of this software and associated documentation files (the "Software"),
//to deal in the Software without restriction, including without limitation
//the rights to use, copy, modify, merge, publish, distribute, sublicense,
//and/or sell copies of the Software, and to permit persons to whom the
//Software is furnished to do so, subject to the following conditions:
//
//The above copyright notice and this permission notice shall be included in
//all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
//THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
//FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
//DEALINGS IN THE SOFTWARE.
//
//
//----------------------------------------------------------------------------
//
//
//SteerLibraryMixin
//
//This mixin (class with templated superclass) adds the "steering library"
//functionality to a given base class.  SteerLibraryMixin assumes its base
//class supports the AbstractVehicle interface.
//
//02-06-03 cwr: create mixin (from "SteerMass")
//06-03-02 cwr: removed TS dependencies
//11-21-01 cwr: created
//
//
//----------------------------------------------------------------------------

public class SteerVehicle extends SimpleVehicle
{
    // Constructor: initializes state
    public SteerVehicle()
    {
        super();
        // set inital state
        reset();
    }

    // reset state
    public void reset()
    {
        super.reset ();
        
        // initial state of wander behavior
        WanderSide = 0;
        WanderUp = 0;

        // default to non-gaudyPursuitAnnotation
        gaudyPursuitAnnotation = false;
    }

    public float WanderSide;
    public float WanderUp;

    public Vector3f hisPositionAtNearestApproach;
    public Vector3f ourPositionAtNearestApproach;

    public boolean gaudyPursuitAnnotation;

    // ----------------------------------------------------------- utilities
    // XXX these belong somewhere besides the steering library
    // XXX above AbstractVehicle, below SimpleVehicle
    // XXX ("utility vehicle"?)

    // xxx cwr experimental 9-9-02 -- names OK?
    public boolean isAhead(Vector3f target)
    {
        return isAhead(target, 0.707f);
    }

    public boolean isAside(Vector3f target)
    {
        return isAside(target, 0.707f);
    }

    public boolean isBehind(Vector3f target)
    {
        return isBehind(target, -0.707f);
    }

    public boolean isAhead(Vector3f target, float cosThreshold)
    {
        // final Vector3f targetDirection = (target - position()).normalize();
        final Vector3f targetDirection = new Vector3f (target); targetDirection.sub (position()); targetDirection.normalize(); 
        return forward().dot(targetDirection) > cosThreshold;
    }

    public boolean isAside(Vector3f target, float cosThreshold)
    {
        // final Vector3f targetDirection = (target - position()).normalize();
        final Vector3f targetDirection = new Vector3f (target); targetDirection.sub (position()); targetDirection.normalize(); 
        final float dp = forward().dot(targetDirection);
        return (dp < cosThreshold) && (dp > -cosThreshold);
    }

    public boolean isBehind(Vector3f target, float cosThreshold)
    {
        // final Vector3f targetDirection = (target - position()).normalize();
        final Vector3f targetDirection = new Vector3f (target); targetDirection.sub (position()); targetDirection.normalize(); 
        return forward().dot(targetDirection) < cosThreshold;
    }

    public static float scalarRandomWalk (float initial, float walkspeed, float min, float max)
    {
        final float next = initial + ((((float) Math.random() * 2) - 1) * walkspeed);
        if (next < min) return min;
        if (next > max) return max;
        return next;
    }
    
    public Vector3f xxxsteerForWander(float dt) {
    	
        WanderSide = scalarRandomWalk(WanderSide, dt, -maxForce() * 0.3f, maxForce() * 0.3f);
        WanderUp = scalarRandomWalk(WanderUp, dt, -maxForce() * 0.3f, maxForce() * 0.3f);

        // return a pure lateral steering vector: (+/-Side) + (+/-Up)
        // return (side() * WanderSide) + (up() * WanderUp);
        Vector3f v1 = new Vector3f (side()); v1.scale(WanderSide);
        Vector3f v2 = new Vector3f (forward()); v2.scale (WanderUp);
        //v1.add(v2);
        return v1;
    }

    public Vector3f steerForWander(float dt)
    {
        // random walk WanderSide and WanderUp between -1 and +1
        final float speed = 12 * dt; // maybe this (12) should be an argument?
        WanderSide = scalarRandomWalk(WanderSide, speed, -10, +10);
        //WanderUp = scalarRandomWalk(WanderUp, speed, -1, +1);

        // return a pure lateral steering vector: (+/-Side) + (+/-Up)
        // return (side() * WanderSide) + (up() * WanderUp);
        Vector3f v1 = new Vector3f (side()); v1.scale(WanderSide);
        //Vector3f v2 = new Vector3f (up()); v2.scale (WanderUp);
        //v1.add(v2);
        return v1;
    }

    public Vector3f steerForSeek(Vector3f target)
    {
        // final Vector3f desiredVelocity = target - position();
        // return desiredVelocity - velocity();
        final Vector3f desiredVelocity = new Vector3f (target); desiredVelocity.sub (position()); 
        
        desiredVelocity.sub(velocity());
        return desiredVelocity;        
    }

    public Vector3f steerForFlee(Vector3f target)
    {
        // final Vector3f desiredVelocity = position - target;
        // return desiredVelocity - velocity();
        final Vector3f desiredVelocity = new Vector3f (position()); desiredVelocity.sub (target); 
        desiredVelocity.sub (velocity());
        return desiredVelocity;        
    }
    
    public Vector3f xxxsteerForArrive(Vector3f target, float slowingDistance) {
        
        Vector3f offset = new Vector3f(target);
        offset.sub(this.position());
        
        float distance = offset.length();
        
        float rampedSpeed = this.maxSpeed() * distance / slowingDistance;
        
        if ( rampedSpeed > maxSpeed() ) rampedSpeed = maxSpeed();
        
        offset.scale( rampedSpeed / distance );
        offset.sub(velocity());
        
        return offset;
    }
    
    public Vector3f steerForArrive(Vector3f target, float slowingDistance, float stopDistance)
    {
    	Vector3f desiredVelocity = xxxsteerForSeek(target);
    	
    	Vector3f localTarget = localizePosition(target);
    	float distance = localTarget.length();

    	Vector3f brakingForce = new Vector3f(ZERO);
    	
    	if ( distance < slowingDistance  && distance > 0) {
    		
    			brakingForce = new Vector3f(forward());
    			brakingForce.scale( -2f * mass() * speed() * speed() / distance );
    		
    			desiredVelocity.add(brakingForce);
    	}
    	if ( Math.abs(distance) < stopDistance ) {
    		
    		return ZERO;
    	}

    	
    	return desiredVelocity;
    }

    public Vector3f xxxsteerForFlee(Vector3f target)
    {
        // final Vector3f offset = position() - target;
        final Vector3f offset = new Vector3f (position()); offset.sub (target); 
        final Vector3f desiredVelocity = Vector3fUtil.truncateLength(offset, maxSpeed()); //xxxnew
        desiredVelocity.sub (velocity());
        return desiredVelocity;        
    }

    public Vector3f xxxsteerForSeek(Vector3f target)
    {
        // final Vector3f desiredVelocity = target - position();
        // return desiredVelocity - velocity();
        final Vector3f offset = new Vector3f (target); offset.sub (position()); 
        final Vector3f desiredVelocity = Vector3fUtil.truncateLength(offset, maxSpeed()); //xxxnew
        desiredVelocity.sub (velocity());
        return desiredVelocity;        
    }

    public Vector3f steerForPathFollowing(float predictionTime, Pathway path)
    {
        // predict our future position
        final Vector3f futurePosition = predictFuturePosition(predictionTime);

        // find the point on the path nearest the predicted future position
        Vector3f onPath = new Vector3f (), tangent = new Vector3f ();
        float distance = path.mapPointToPath(futurePosition, onPath, tangent);
        
        boolean withinPath = distance + radius() < path.getRadius(); 

//        annotationForPathFollowing(futurePosition, onPath, withinPath);

        if (withinPath)
        {
            // our predicted future position was in the path,
            // return zero steering.
            return LocalSpace.ZERO;
        }
        else
        {
            // our predicted future position was outside the path, need to
            // steer towards it.  Decide whether we are going up or down
            // the path.  Choose a target point on the path, somewhat
            // "ahead: of our current position.
            // final Vector3f nextPosition = position() + forward();
            final Vector3f nextPosition = new Vector3f (position()); 
            nextPosition.add (forward());
            
            float thisPathDistance = path.mapPointToPathDistance(position());
            float nextPathDistance = path.mapPointToPathDistance(nextPosition);
            // xxx!:
            float nudge = (thisPathDistance < nextPathDistance) ? 20 : -20;
            Vector3f target = path.mapPathDistanceToPoint(thisPathDistance + nudge);
            return xxxsteerForSeek(target);
        }
    }

    public Vector3f steerToAvoidObstacle(Obstacle obstacle, float minTimeToCollision)
    {
        final Vector3f avoidance = obstacle.steerToAvoid(this, minTimeToCollision);

        // XXX more annotation modularity problems (assumes spherical obstacle)
        if (avoidance != LocalSpace.ZERO)
        {
//            annotationForAvoidSphere(minTimeToCollision * speed());
        }

        return avoidance;
    }
    
    public Vector3f steerToAvoidObstacles (List obstacles, float minTimeToCollision)
    {
        if (obstacles.isEmpty()) return ZERO;
        
        Vector3f avoidance = new Vector3f ();
        
        for (Iterator i = obstacles.iterator(); i.hasNext(); )
        {
            avoidance.add (steerToAvoidObstacle ((Obstacle) i.next(), minTimeToCollision));
        }
        
        return avoidance;
    }
    
    public Vector3f xxxsteerToAvoidNeighbors(float minTimeToCollision, List others)
    {
        // first priority is to prevent immediate interpenetration
        final Vector3f separation = steerToAvoidCloseNeighbors(0, others);
        if (separation != ZERO)
            return separation;

        // otherwise, go on to consider potential future collisions
        float steer = 0;
        AbstractVehicle threat = null;

        // Time (in seconds) until the most immediate collision threat found
        // so far.  Initial value is a threshold: don't look more than this
        // many frames into the future.
        float minTime = minTimeToCollision;

        // xxx solely for annotation
        Vector3f xxxThreatPositionAtNearestApproach = null;
        Vector3f xxxOurPositionAtNearestApproach = null;

        // for each of the other vehicles, determine which (if any)
        // pose the most immediate threat of collision.
        for (Iterator i = others.iterator(); i.hasNext(); )
        {
            AbstractVehicle other = (AbstractVehicle) i.next();
            if (other != this)
            {
                // avoid when future positions are this close (or less)
                final float collisionDangerThreshold = radius() * 3;

                // predicted time until nearest approach of "this" and "other"
                final float time = predictNearestApproachTime(other);

                // If the time is in the future, sooner than any other
                // threatened collision...
                if ((time >= 0) && (time < minTime))
                {
                    // if the two will be close enough to collide,
                    // make a note of it
                    if (computeNearestApproachPositions(other, time) < collisionDangerThreshold)
                    {
                        minTime = time;
                        threat = other;
                        xxxThreatPositionAtNearestApproach = hisPositionAtNearestApproach;
                        xxxOurPositionAtNearestApproach = ourPositionAtNearestApproach;
                    }
                }
            }
        }

        // if a potential collision was found, compute steering to avoid
        if (threat != null)
        {
        	Vector3f offset = new Vector3f(ourPositionAtNearestApproach);
        	offset.sub(hisPositionAtNearestApproach);
        	
        	offset.normalize();
        	
        	float avoidForce = (1 - minTime / minTimeToCollision) * maxForce(); 
        	
        	offset.scale(avoidForce);
        	
        	return offset;
        }
        
        // return side() * steer;
        return ZERO;
    }

    public Vector3f steerToAvoidNeighbors(float minTimeToCollision, List others)
    {
        // first priority is to prevent immediate interpenetration
        final Vector3f separation = steerToAvoidCloseNeighbors(0, others);
        if (separation != ZERO)
            return separation;

        // otherwise, go on to consider potential future collisions
        float steer = 0;
        AbstractVehicle threat = null;

        // Time (in seconds) until the most immediate collision threat found
        // so far.  Initial value is a threshold: don't look more than this
        // many frames into the future.
        float minTime = minTimeToCollision;

        // xxx solely for annotation
        Vector3f xxxThreatPositionAtNearestApproach = null;
        Vector3f xxxOurPositionAtNearestApproach = null;

        // for each of the other vehicles, determine which (if any)
        // pose the most immediate threat of collision.
        for (Iterator i = others.iterator(); i.hasNext(); )
        {
            AbstractVehicle other = (AbstractVehicle) i.next();
            if (other != this)
            {
                // avoid when future positions are this close (or less)
                final float collisionDangerThreshold = radius() * 3;

                // predicted time until nearest approach of "this" and "other"
                final float time = predictNearestApproachTime(other);

                // If the time is in the future, sooner than any other
                // threatened collision...
                if ((time >= 0) && (time < minTime))
                {
                    // if the two will be close enough to collide,
                    // make a note of it
                    if (computeNearestApproachPositions(other, time) < collisionDangerThreshold)
                    {
                        minTime = time;
                        threat = other;
                        xxxThreatPositionAtNearestApproach = hisPositionAtNearestApproach;
                        xxxOurPositionAtNearestApproach = ourPositionAtNearestApproach;
                    }
                }
            }
        }

        // if a potential collision was found, compute steering to avoid
        if (threat != null)
        {
            // parallel: +1, perpendicular: 0, anti-parallel: -1
            float parallelness = forward().dot(threat.forward());
            float angle = 0.707f;

            if (parallelness < -angle)
            {
                // anti-parallel "head on" paths:
                // steer away from future threat position
                // Vector3f offset = xxxThreatPositionAtNearestApproach - position();
                Vector3f offset = new Vector3f (xxxThreatPositionAtNearestApproach); offset.sub (position());
                float sideDot = offset.dot(side());
                steer = (sideDot > 0) ? -1 : 1;
            }
            else
            {
                if (parallelness > angle)
                {
                    // parallel paths: steer away from threat
                    // Vector3f offset = threat.position() - position();
                    Vector3f offset = new Vector3f (threat.position()); offset.sub (position());
                    float sideDot = offset.dot(side());
                    steer = (sideDot > 0) ? -1 : 1;
                }
                else
                {
                    // perpendicular paths: steer behind threat
                    // (only the slower of the two does this)
                    if (threat.speed() <= speed())
                    {
                        float sideDot = side().dot(threat.velocity());
                        steer = (sideDot > 0) ? -1 : 1;
                    }
                }
            }

//            annotationForAvoidNeighbor(threat, steer, xxxOurPositionAtNearestApproach, xxxThreatPositionAtNearestApproach);
//            Vector3f offset = new Vector3f(position());
//            offset.sub(threat.position());
//            
//            float avoidForce = 1 / offset.length();
//            if( avoidForce > maxForce() ) avoidForce = maxForce();
            
            float avoidForce = (1 - minTime / minTimeToCollision) * maxForce(); 
            
            Vector3f v = new Vector3f (side()); v.scale (steer * avoidForce); return v;
        }
        
        // return side() * steer;
        return ZERO;
    }

    public float predictNearestApproachTime(AbstractVehicle other)
    {
        // imagine we are at the origin with no velocity,
        // compute the relative velocity of the other vehicle
        final Vector3f myVelocity = velocity();
        final Vector3f otherVelocity = other.velocity();
        // final Vector3f relVelocity = otherVelocity - myVelocity;
        final Vector3f relVelocity = new Vector3f (otherVelocity); relVelocity.sub (myVelocity);
        final float relSpeed = relVelocity.length();

        // Now consider the path of the other vehicle in this relative
        // space, a line defined by the relative position and velocity.
        // The distance from the origin (our vehicle) to that line is
        // the nearest approach.

        // Take the unit tangent along the other vehicle's path
        // final Vector3f relTangent = relVelocity / relSpeed;
        final Vector3f relTangent = new Vector3f (relVelocity); relTangent.scale (1/relSpeed);

        // find distance from its path to origin (compute offset from
        // other to us, find length of projection onto path)
        // final Vector3f relPosition = position() - other.position();
        final Vector3f relPosition = new Vector3f (position()); relPosition.sub (other.position());
        final float projection = relTangent.dot(relPosition);

        return projection / relSpeed;
    }

    public float computeNearestApproachPositions(AbstractVehicle other, float time)
    {
        // final Vector3f myTravel = forward() * speed() * time;
        final Vector3f myTravel = new Vector3f (forward()); myTravel.scale (speed() * time);
        // final Vector3f otherTravel = other.forward() * other.speed() * time;
        final Vector3f otherTravel = new Vector3f (other.forward()); otherTravel.scale (other.speed() * time);

        // final Vector3f myFinal = position() + myTravel;
        final Vector3f myFinal = new Vector3f (position()); myFinal.add (myTravel);
        // final Vector3f otherFinal = other.position() + otherTravel;
        final Vector3f otherFinal = new Vector3f (other.position()); otherFinal.add (otherTravel);

        // xxx for annotation
        ourPositionAtNearestApproach = myFinal;
        hisPositionAtNearestApproach = otherFinal;

        return Vector3fUtil.distance(myFinal, otherFinal);
    }

    public Vector3f steerToAvoidCloseNeighbors(float minSeparationDistance, List others)
    {
        // for each of the other vehicles...
        for (Iterator i = others.iterator(); i.hasNext(); )
        {
            AbstractVehicle other = (AbstractVehicle) i.next();
            if (other != this)
            {
                final float sumOfRadii = radius() + other.radius();
                final float minCenterToCenter = minSeparationDistance + sumOfRadii;
                // final Vector3f offset = other.position() - position();
                final Vector3f offset = new Vector3f (other.position()); offset.sub (position());
                final float currentDistance = offset.length();

                if (currentDistance < minCenterToCenter)
                {
//                    annotationForAvoidCloseNeighbor(other, minSeparationDistance);
                    offset.negate();
                    return Vector3fUtil.perpendicularComponent(offset, forward());
                }
            }
        }

        // otherwise return zero
        return LocalSpace.ZERO;
    }

    public boolean inBoidNeighborhood (Vector3f queryLocation, float minDistance, float maxDistance, float cosMaxAngle)
//        float distanceWeight,
//        Vector3f unitOffset)
    {
        // final Vector3f offset = queryLocation - position();
        final Vector3f offset = new Vector3f (queryLocation); offset.sub (position());
        final float distance = offset.length();
        offset.scale (1/distance);
        Vector3f unitOffset = offset;
        final float forwardness = forward().dot(unitOffset);

        // conical weighting function (linear falloff)
        final float d = maxDistance - distance; //xxx NAME???
//        distanceWeight = d / maxDistance;

        if (distance < minDistance)
        {
            return true;
        }
        else
        {
            if ((distance < maxDistance) && (forwardness > cosMaxAngle))
            {
                return true;
            }
            else
            {
//                distanceWeight = 0;
                return false;
            }
        }
    }

    public Vector3f steerForSeparation(float maxDistance, float cosMaxAngle, Vector flock)
    {
        // for each of the other vehicles...
        Vector3f steering = new Vector3f ();
        int neighbors = 0;
        for (Iterator i = flock.iterator(); i.hasNext(); )
        {
            AbstractVehicle other = (AbstractVehicle) i.next();
            if (other != this)
            {
//                float distanceWeight;
//                Vector3f unitOffset;
                if (inBoidNeighborhood(other.position(), radius() * 3, maxDistance, cosMaxAngle))
                {
                    // XXX ------------------------------ do it like in psboids
                    // // repulsive force, diminished with distance
                    // steering += unitOffset * -distanceWeight;
                    // XXX ------------------------------ do it like in psboids
                    // final Vector3f offset = other.position() - position();
                    final Vector3f offset = new Vector3f (other.position()); offset.sub (position());
                    final float distanceSquared = offset.dot(offset);
                    // steering += (offset / -distanceSquared);
                    offset.scale (-1/distanceSquared);
                    steering.add (offset);
                    // XXX ------------------------------ do it like in psboids

                    // count neighbors
                    neighbors++;
                }
            }
        }

        // XXX ------------------------------ do it like in psboids
        //     if (neighbors > 0) steering = steering / neighbors;
        // XXX ------------------------------ do it like in psboids
        if (neighbors > 0)
        {
            // steering = (steering / neighbors).normalize();
            steering.scale (1.0f / neighbors); 
            steering.normalize();
        }
        // XXX ------------------------------ do it like in psboids

        return steering;
    }

    public Vector3f steerForAlignment(float maxDistance, float cosMaxAngle, Vector flock)
    {
        // for each of the other vehicles...
        Vector3f steering = new Vector3f ();
        int neighbors = 0;
        for (Iterator i = flock.iterator(); i.hasNext(); )
        {
            AbstractVehicle other = (AbstractVehicle) i.next();
            if (other != this)
            {
//                float distanceWeight;
//                Vector3f unitOffset;
                if (inBoidNeighborhood(other.position(), radius() * 3, maxDistance, cosMaxAngle))
                {
                    // const Vector3f alignError = forward() - other.forward;
                    // steering += alignError * -distanceWeight;
                    // XXX ------------------------------ do it like in psboids
                    // const Vector3f myVelocity = forward() * speed;
                    // const Vector3f otherVelocity = other.forward() * other.speed;
                    // const Vector3f velocityDifference = myVelocity - otherVelocity;
                    // XXX ------------------------------ do it like in psboids

                    // XXX ------------------------------ do it like in psboids
                    // // corrective force, diminished with distance
                    // steering += velocityDifference * -distanceWeight;
                    // XXX ------------------------------ do it like in psboids
                    steering.add (other.forward());
                    // XXX ------------------------------ do it like in psboids

                    // count neighbors
                    neighbors++;
                }

            }
        }

        // XXX ------------------------------ do it like in psboids
        //     if (neighbors > 0) steering = steering / neighbors;
        // XXX ------------------------------ do it like in psboids
        if (neighbors > 0)
        {
            steering.scale (1.0f / neighbors);
            steering.sub (forward());
            steering.normalize();
        }
        // XXX ------------------------------ do it like in psboids

        return steering;
    }

    public Vector3f steerForCohesion(float maxDistance, float cosMaxAngle, Vector flock)
    {
        // for each of the other vehicles...
        Vector3f steering = new Vector3f ();
        int neighbors = 0;
        for (Iterator i = flock.iterator(); i.hasNext(); )
        {
            AbstractVehicle other = (AbstractVehicle) i.next();
            if (other != this)
            {

//                float distanceWeight;
//                Vector3f unitOffset;
                if (inBoidNeighborhood(other.position(), radius() * 3, maxDistance, cosMaxAngle))
                {
                    // XXX ------------------------------ do it like in psboids
                    // // attractive force, diminished with distance
                    // steering += unitOffset * distanceWeight;
                    // XXX ------------------------------ do it like in psboids
                    steering.add (other.position());
                    // XXX ------------------------------ do it like in psboids

                    // count neighbors
                    neighbors++;
                }
            }
        }

        // XXX ------------------------------ do it like in psboids
        //     if (neighbors > 0) steering = steering / neighbors;
        // XXX ------------------------------ do it like in psboids
        if (neighbors > 0)
        {
            steering.scale (1.0f / neighbors);
            steering.sub (position());
            steering.normalize();
        }
        // XXX ------------------------------ do it like in psboids

        return steering;
    }

    public Vector3f steerForPursuit(AbstractVehicle quarry)
    {
        return steerForPursuit(quarry, Float.MAX_VALUE);
    }

    public Vector3f steerForPursuit(AbstractVehicle quarry, float maxPredictionTime)
    {
        // offset from this to quarry, that distance, unit vector toward quarry
        // final Vector3f offset = quarry.position() - position();
        final Vector3f offset = new Vector3f (quarry.position()); offset.sub (position());
        final float distance = offset.length();        
        // final Vector3f unitOffset = offset / distance;
        offset.scale (1 / distance);
        final Vector3f unitOffset = offset;

        // how parallel are the paths of "this" and the quarry
        // (1 means parallel, 0 is pependicular, -1 is anti-parallel)
        final float parallelness = forward().dot(quarry.forward());

        // how "forward" is the direction to the quarry
        // (1 means dead ahead, 0 is directly to the side, -1 is straight back)
        final float forwardness = forward().dot(unitOffset);

        final float directTravelTime = distance / speed();
        final int f = intervalComparison(forwardness, -0.707f, 0.707f);
        final int p = intervalComparison(parallelness, -0.707f, 0.707f);

        float timeFactor = 0; // to be filled in below
        Color3f color; // to be filled in below (xxx just for debugging)

        // Break the pursuit into nine cases, the cross product of the
        // quarry being [ahead, aside, or behind] us and heading
        // [parallel, perpendicular, or anti-parallel] to us.
        switch (f)
        {
            case +1 :
                switch (p)
                {
                    case +1 : // ahead, parallel
                        timeFactor = 4;
                        color = Annotation.gBlack;
                        break;
                    case 0 : // ahead, perpendicular
                        timeFactor = 1.8f;
                        color = Annotation.gGray50;
                        break;
                    case -1 : // ahead, anti-parallel
                        timeFactor = 0.85f;
                        color = Annotation.gWhite;
                        break;
                }
                break;
            case 0 :
                switch (p)
                {
                    case +1 : // aside, parallel
                        timeFactor = 1;
                        color = Annotation.gRed;
                        break;
                    case 0 : // aside, perpendicular
                        timeFactor = 0.8f;
                        color = Annotation.gYellow;
                        break;
                    case -1 : // aside, anti-parallel
                        timeFactor = 4;
                        color = Annotation.gGreen;
                        break;
                }
                break;
            case -1 :
                switch (p)
                {
                    case +1 : // behind, parallel
                        timeFactor = 0.5f;
                        color = Annotation.gCyan;
                        break;
                    case 0 : // behind, perpendicular
                        timeFactor = 2;
                        color = Annotation.gBlue;
                        break;
                    case -1 : // behind, anti-parallel
                        timeFactor = 2;
                        color = Annotation.gMagenta;
                        break;
                }
                break;
        }

        // estimated time until intercept of quarry
        final float et = directTravelTime * timeFactor;

        // xxx experiment, if kept, this limit should be an argument
        final float etl = (et > maxPredictionTime) ? maxPredictionTime : et;

        // estimated position of quarry at intercept
        final Vector3f target = quarry.predictFuturePosition(etl);

        // annotation
//        annotationLine(position(), target, gaudyPursuitAnnotation ? color : Annotation.gGray40);

        return steerForSeek(target);
    }

    public Vector3f steerForEvasion(AbstractVehicle menace, float maxPredictionTime)
    {
        // offset from this to menace, that distance, unit vector toward menace
        // final Vector3f offset = menace.position - position;
        final Vector3f offset = new Vector3f (menace.position()); offset.sub (position());
        final float distance = offset.length();

        final float roughTime = distance / menace.speed();
        final float predictionTime = ((roughTime > maxPredictionTime) ? maxPredictionTime : roughTime);

        final Vector3f target = menace.predictFuturePosition(predictionTime);

        return steerForFlee(target);
    }

    public Vector3f steerForTargetSpeed(float targetSpeed)
    {
        float speedError = targetSpeed - speed();

        if (speedError > +maxForce())
            speedError = +maxForce();
        if (speedError < -maxForce())
            speedError = -maxForce();

        // return forward() * speedError;
        Vector3f v = new Vector3f (forward()); v.scale (speedError); return v;
    }

    public void findNextIntersectionWithSphere(SphericalObstacle obs, PathIntersection intersection)
    {
        float b, c, d, p, q, s;
        Vector3f lc;

        // initialize pathIntersection object
        intersection.intersect = false;
        intersection.obstacle = obs;

        // find "local center" (lc) of sphere in boid's coordinate space
        lc = localizePosition(obs.center);

        // computer line-sphere intersection parameters
        b = -2 * lc.z;
        c = ((lc.x) * (lc.x)) + ((lc.y) * (lc.y)) + ((lc.z) * (lc.z)) - ((obs.radius + radius()) * (obs.radius + radius()));
        d = (b * b) - (4 * c);

        // when the path does not intersect the sphere
        if (d < 0)
            return;

        // otherwise, the path intersects the sphere in two points with
        // parametric coordinates of "p" and "q".
        // (If "d" is zero the two points are coincident, the path is tangent)
        s = (float) Math.sqrt(d);
        p = (-b + s) / 2;
        q = (-b - s) / 2;

        // both intersections are behind us, so no potential collisions
        if ((p < 0) && (q < 0))
            return;

        // at least one intersection is in front of us
        intersection.intersect = true;
        intersection.distance = ((p > 0) && (q > 0)) ?
            // both intersections are in front of us, find nearest one
             ((p < q) ? p : q) :
            // otherwise only one intersections is in front, select it
             ((p > 0) ? p : q);
        return;
    }

//    public void annotationForPathFollowing(Vector3f futurePosition, Vector3f onPath, boolean withinPath)
//    {
//        final Vector3f yellow = new Vector3f(1, 1, 0);
//        final Vector3f lightOrange = new Vector3f(1.0f, 0.5f, 0.0f);
//        final Vector3f darkOrange = new Vector3f(0.5f, 0.2f, 0.0f);
//        final Vector3f orange = new Vector3f(withinPath ? darkOrange : lightOrange);
//
//        annotationLine(futurePosition, position(), yellow);
//        annotationLine(futurePosition, onPath, orange);
//    }

//    public void annotationForAvoidSphere(float minDistanceToCollision)
//    {
//        final Vector3f boxSide = side() * radius();
//        final Vector3f boxFront = forward() * minDistanceToCollision;
//        final Vector3f FR = position() + boxFront - boxSide;
//        final Vector3f FL = position() + boxFront + boxSide;
//        final Vector3f BR = position() - boxSide;
//        final Vector3f BL = position() + boxSide;
//        final Vector3f white = new Vector3f(1, 1, 1);
//        annotationLine(FR, FL, white);
//        annotationLine(FL, BL, white);
//        annotationLine(BL, BR, white);
//        annotationLine(BR, FR, white);
//    }

//    public void annotationForAvoidNeighbor(AbstractVehicle threat, float steer, Vector3f ourFuture, Vector3f threatFuture)
//    {
//        final Vector3f red = new Vector3f(1, 0, 0);
//        final Vector3f cyan = new Vector3f(0, 1, 1);
//        final Vector3f white = new Vector3f(1, 1, 1);
//        final Vector3f magenta = new Vector3f(1, 0, 1);
//        annotationLine(position(), threat.position(), magenta);
//        annotationLine(position(), position() + (side() * steer * 3), cyan);
//        annotationLine(position(), ourFuture, white);
//        annotationLine(threat.position(), threatFuture, white);
//        annotationLine(ourFuture, threatFuture, red);
//    }

//    public void annotationForAvoidCloseNeighbor(AbstractVehicle other, float additionalDistance)
//    {
//        // draw the word "Ouch!" above colliding vehicles
//        final float headOn = forward().dot(other.forward()) < 0;
//        final Vector3f green = new Vector3f(0.4f, 0.8f, 0.1f);
//        final Vector3f red = new Vector3f(1, 0.1f, 0);
//        final Vector3f color = headOn ? red : green;
//        final String string = headOn ? "OUCH!" : "pardon me";
//        final Vector3f location = position() + new Vector3f(0, 0.5, 0);
//        draw2dTextAt3dLocation(string, location, color);
//    }

    
        // xxx cwr 9-6-02 temporary to support old code
    class PathIntersection
    {
        boolean intersect;
        float distance;
        Vector3f surfacePoint;
        Vector3f surfaceNormal;
        SphericalObstacle obstacle;
    };
}
